import os
import torch
import datetime
import numpy as np
from torch.utils.data import DataLoader
from ..models import WTM
from .dataset import make_data_contrastive_square, make_data_contrastive_hexagon, make_data_contrastive_others, make_test_data_contrastive, train_Dataset_contrastive,  test_Dataset_contrastive
import random
import pandas as pd



def setup_seed(seed):
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    random.seed(seed)


def normalize_expr(data):
    data = data / (data.sum(axis=1, keepdims=True) + 1e-10)

    return data


def collate_fn(batch):
    spots, pos_samples, neg_samples, pos_num, neg_num = zip(*batch)

    return torch.cat(spots, dim=0), torch.cat(pos_samples, dim=0), torch.cat(neg_samples, dim=0), torch.tensor(pos_num), torch.tensor(neg_num)


def deconvolution(df_data, cell_type_num, spatial_mode, save_dir_name, pos_rate=0.05, neg_rate=0.05, alpha=0.2, margin=0.05, batch_size=64, num_epoch=500, num_workers=0, device_id=None, random_seed=0):
    setup_seed(random_seed)
    if device_id != None:
        device = torch.device('cuda:{}'.format(device_id))
    else:
        device = torch.device('cpu')
        print('It is recommended that you use GPU for acceleration, otherwise the running time will be very long!')

    print('\nData organization...')
    # df_expr = pd.read_csv(data_dir + 'spot_pos_and_expr.csv')
    df_expr = df_data
    spatial_fea_num = df_expr.shape[1] - 3
    df_expr.iloc[:, 3:] = normalize_expr(df_expr.iloc[:, 3:].to_numpy())

    if spatial_mode == 'square':
        train_data = make_data_contrastive_square(df_expr, pos_rate=pos_rate, neg_rate=neg_rate)
    elif spatial_mode == 'hexagon':
        train_data = make_data_contrastive_hexagon(df_expr, pos_rate=pos_rate, neg_rate=neg_rate)
    else:
        train_data = make_data_contrastive_others(df_expr, pos_rate=pos_rate, neg_rate=neg_rate)
    print('Total_spot_num:{}'.format(len(train_data)))
    test_data, pos_all = make_test_data_contrastive(df_expr)

    train_set = train_Dataset_contrastive(data=train_data)
    test_set = test_Dataset_contrastive(data=test_data)
    train_dataloader = DataLoader(train_set, batch_size=batch_size, shuffle=True, num_workers=num_workers, pin_memory=True, collate_fn=collate_fn)
    test_dataloader = DataLoader(test_set, batch_size=batch_size, shuffle=False, num_workers=num_workers, pin_memory=True)
    print('Data organization completed.')

    begin_time = datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
    print('begin time:{}'.format(begin_time))

    print('\nDeconvolution begin...')
    model = WTM(bow_dim=spatial_fea_num, n_topic=cell_type_num, device=device, dropout=0.5, alpha=alpha)
    train_save_dir = "results_save/{}_{}/model_save/".format(save_dir_name, begin_time)
    os.makedirs(train_save_dir, exist_ok=True)
    model.train(train_dataloader=train_dataloader, num_epochs=num_epoch, save_dir=train_save_dir, margin=margin)
    embeds, recon = model.get_embed(dataloader=test_dataloader)
    spatial_gene_list = df_expr.columns[3:].tolist()
    top_gene_ids, top_gene_names, beta = model.get_topic_top_words(gene_names_list=spatial_gene_list, top_k=5)
    print('Deconvolution completed')

    df_deconvolution_results = pd.DataFrame(embeds, columns=[i for i in range(embeds.shape[1])])
    df_deconvolution_results = pd.concat([df_expr.iloc[:, :3], df_deconvolution_results], axis=1)
    df_beta = pd.DataFrame(beta, columns=spatial_gene_list, index=[i for i in range(embeds.shape[1])])
    df_beta = df_beta.reset_index()

    save_dir = "results_save/{}_{}/prediction_save/".format(save_dir_name, begin_time)
    os.makedirs(save_dir, exist_ok=True)
    df_deconvolution_results.to_csv(save_dir + 'pred.csv', index=False)
    df_beta.to_csv(save_dir + 'beta.csv', index=False)

    return df_deconvolution_results, df_beta, test_dataloader


def deconvolution_multi_ctn(df_data, cell_type_num, spatial_mode, save_dir_name, pos_rate=0.05, neg_rate=0.05, alpha=0.2, margin=0.05, batch_size=64, num_epoch=500, device_id=None, random_seed=0, num_workers=0):
    setup_seed(random_seed)
    if device_id != None:
        device = torch.device('cuda:{}'.format(device_id))
    else:
        device = torch.device('cpu')
        print('It is recommended that you use GPU for acceleration, otherwise the running time will be very long!')

    print('\nData organization...')

    df_expr = df_data
    spatial_fea_num = df_expr.shape[1] - 3
    df_expr.iloc[:, 3:] = normalize_expr(df_expr.iloc[:, 3:].to_numpy())

    if spatial_mode == 'square':
        train_data = make_data_contrastive_square(df_expr, pos_rate=pos_rate, neg_rate=neg_rate)
    elif spatial_mode == 'hexagon':
        train_data = make_data_contrastive_hexagon(df_expr, pos_rate=pos_rate, neg_rate=neg_rate)
    else:
        train_data = make_data_contrastive_others(df_expr, pos_rate=pos_rate, neg_rate=neg_rate)
    print('Total_spot_num:{}'.format(len(train_data)))
    test_data, pos_all = make_test_data_contrastive(df_expr)

    train_set = train_Dataset_contrastive(data=train_data)
    test_set = test_Dataset_contrastive(data=test_data)
    train_dataloader = DataLoader(train_set, batch_size=batch_size, shuffle=True, num_workers=num_workers, pin_memory=True, collate_fn=collate_fn)
    test_dataloader = DataLoader(test_set, batch_size=batch_size, shuffle=False, num_workers=num_workers, pin_memory=True)
    print('Data organization completed.')

    begin_time = datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
    print('begin time:{}'.format(begin_time))

    parameter_adjustment = []
    begin_time = datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
    for cell_type_num_i in cell_type_num:
        setup_seed(random_seed)
        print(f'\nCell type number = {cell_type_num_i}, begin deconvolution....')
        os.makedirs("results_save/{}_{}/model_save/cell_type_num_{}".format(save_dir_name, begin_time, cell_type_num_i), exist_ok=True)
        model = WTM(bow_dim=spatial_fea_num, n_topic=cell_type_num_i, device=device, dropout=0.5, alpha=alpha)
        model.train(train_dataloader=train_dataloader, num_epochs=num_epoch, save_dir="results_save/{}_{}/model_save/cell_type_num_{}".format(save_dir_name, begin_time, cell_type_num_i), margin=margin)

        embeds, recons = model.get_embed(dataloader=test_dataloader)
        top_gene_ids, top_gene_names, beta = model.get_topic_top_words(gene_names_list=df_expr.columns.tolist()[3:], top_k=5)
        rare_num, rare_cell_type = model.calc_rare_cell_type_num(embeds, thres=0.05)
        ppl = model.calc_RD(test_data, recons)
        os.makedirs("results_save/{}_{}/prediction_save/cell_type_num_{}".format(save_dir_name, begin_time, cell_type_num_i), exist_ok=True)
        dict_i = {'cell_type_num': cell_type_num_i, 'rare_cell_type_num': rare_num, 'ppl': ppl}

        parameter_adjustment.append(dict_i)

        df = pd.DataFrame(parameter_adjustment)
        save_dir = 'results_save/{}_{}/'.format(save_dir_name, begin_time)
        df.to_csv(save_dir + '{}.csv'.format(save_dir_name), index=False)
        df_pred = pd.DataFrame(embeds.squeeze())
        df_pred.to_csv(save_dir + 'prediction_save/cell_type_num_{}/pred.csv'.format(cell_type_num_i), index=False)
        df_beta = pd.DataFrame(beta, columns=df_expr.columns[3:])
        df_beta.to_csv(save_dir + 'prediction_save/cell_type_num_{}/beta.csv'.format(cell_type_num_i), index=False)

    return df, test_dataloader



import os
import rpy2.robjects as robjects


def run_cellchat(count_matrix, pos_data, barcodes, meta, scale_data_dir, plot_dir, contact_range=100, spot_size=65, species='human'):
    robjects.globalenv['count_matrix'] = count_matrix
    robjects.globalenv['pos_data'] = pos_data
    robjects.globalenv['barcodes'] = barcodes
    robjects.globalenv['meta'] = meta
    robjects.globalenv['scale_data_dir'] = scale_data_dir
    robjects.globalenv['plot_dir'] = plot_dir
    robjects.globalenv['contact_range'] = contact_range
    robjects.globalenv['spot.size'] = spot_size
    robjects.globalenv['species'] = species
    
    os.makedirs(plot_dir, exist_ok=True)

    r_code = """
    library(CellChat)
    library(patchwork)
    library(Seurat)
    library(Matrix)
    options(stringsAsFactors = FALSE)
    
    print('Preprocess data...')
    scalefactors <- jsonlite::fromJSON(txt = file.path(scale_data_dir, 'scalefactors_json.json'))
    meta$samples <- factor(meta$samples)
    rownames(count_matrix) <- barcodes
    pos_data$barcode <- barcodes
    count_matrix <- t(count_matrix)
    rownames(pos_data) <- pos_data$barcode

    count_matrix <- count_matrix[, rownames(meta)]
    pos_data <- pos_data[rownames(meta), ]

    data.input <- normalizeData(count_matrix)
    spatial.locs <- pos_data[, c("x_img", "y_img")]
    colnames(spatial.locs) <- c("imagerow", "imagecol")

    conversion.factor = spot.size/scalefactors$spot_diameter_fullres
    spatial.factors = data.frame(ratio=conversion.factor, tol=spot.size/2)
    d.spatial <- computeCellDistance(coordinates=spatial.locs, ratio=spatial.factors$ratio, tol=spatial.factors$tol)
    
    if (species == "human") {
        CellChatDB <- CellChatDB.human
    } else if (species == "mouse") {
        CellChatDB <- CellChatDB.mouse
    } else {
        stop("Unsupported species. Please use 'human' or 'mouse'")
    }
   
    print('Create Cellchat data...')
    cellchat <- createCellChat(object=data.input, meta=meta, group.by="labels", datatype="spatial", coordinates=spatial.locs, spatial.factors=spatial.factors)
    cellchat@DB <- CellChatDB

    cellchat <- subsetData(cellchat)
    future::plan("multisession", workers=4)
    cellchat <- identifyOverExpressedGenes(cellchat)
    cellchat <- identifyOverExpressedInteractions(cellchat, variable.both=F)
    
    print('Compute communication probabilities...')
    cellchat <- computeCommunProb(cellchat, type="truncatedMean", trim=0.1, distance.use=TRUE, interaction.range=250, scale.distance=0.01, contact.dependent=TRUE, contact.range=contact_range)
    cellchat <- filterCommunication(cellchat, min.cells=20)
    df.net <- subsetCommunication(cellchat)
    cellchat <- computeCommunProbPathway(cellchat)
    cellchat <- aggregateNet(cellchat)

    groupSize <- as.numeric(table(cellchat@idents))
    
    print('Plot...')
    png(file.path(plot_dir, 'interaction_circle.jpg'), width=2000, height=2000, res=300)
    par(mfrow = c(1,1), xpd=TRUE)
    netVisual_circle(cellchat@net$weight,
        vertex.weight = rowSums(cellchat@net$weight),
        weight.scale = TRUE, label.edge = FALSE,
        title.name = 'Interaction weights/strength')
    dev.off()
    
   
    png(file.path(plot_dir, 'CLDN_interaction_circle.jpg'), width=2000, height=2000, res=300)
    pathways.show <- c("CLDN")
    par(mfrow=c(1,2), xpd = TRUE)
    netVisual_aggregate(cellchat, signaling = pathways.show, layout = "circle", sources.use = c(1, 2, 4), targets.use = c(5, 6))
    netVisual_aggregate(cellchat, signaling = pathways.show, layout = "circle", sources.use = c(5, 6), targets.use = c(1, 2, 4))
    dev.off()
    
    
    png(file.path(plot_dir, 'CypA_interaction_circle.jpg'), width=2000, height=2000, res=300)
    pathways.show <- c("CypA")
    par(mfrow=c(1,2), xpd = TRUE)
    netVisual_aggregate(cellchat, signaling = pathways.show, layout = "circle", sources.use = c(1, 2, 4), targets.use = c(5, 6))
    netVisual_aggregate(cellchat, signaling = pathways.show, layout = "circle", sources.use = c(5, 6), targets.use = c(1, 2, 4))
    dev.off()
    """
    robjects.r('options(device = "png")')
    robjects.r('options(width=100)')
    robjects.r(r_code)


import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as init
import time
import multiprocessing as mp
from torch.nn.utils.rnn import pad_sequence


class WAE(nn.Module):
    def __init__(self, encode_dims=[2000, 1024, 512, 20], decode_dims=[20, 512, 2000], dropout=0.0, device=None):
        super(WAE, self).__init__()

        self.device = device

        # encoder_linear
        self.enc_linear = nn.Sequential(
            nn.Linear(encode_dims[0], 32),
            nn.BatchNorm1d(32),
            nn.LeakyReLU(),
            nn.Dropout(p=dropout),
            nn.Linear(32, encode_dims[-1]),
        )

        self.z_dim = encode_dims[-1]

        self.decoder = nn.Sequential(
            nn.Linear(encode_dims[-1], decode_dims[2]),
        )

        self.celltype_weight = nn.Parameter(torch.ones(encode_dims[-1]))

    def encode(self, spot_fea):
        hid = self.enc_linear(spot_fea)

        return hid

    def decode(self, theta):
        hid = self.decoder(theta)
        return hid

    def forward(self, spot_fea):
        z = self.encode(spot_fea)
        theta = F.softmax(z, dim=1)
        x_reconst = self.decode(theta.squeeze())
        return x_reconst, theta.squeeze(), F.softmax(self.celltype_weight, dim=0)

    def initialize(self):
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.xavier_uniform_(m.weight.data)
                # nn.init.kaiming_normal_(m.weight.data)

    def sample(self, alpha, batch_size=256):
        z_true = np.random.dirichlet(alpha, size=batch_size)
        z_true = torch.from_numpy(z_true).float()
        return z_true

    def mmd_loss(self, x, y, device, t=0.1, kernel='diffusion'):
        '''
        computes the mmd loss with information diffusion kernel
        :param x: batch_size * latent dimension
        :param y:
        :param t:
        :return:
        '''
        eps = 1e-6
        n = x.shape[0]
        if kernel == 'tv':
            sum_xx = torch.zeros(1).to(device)
            for i in range(n):
                for j in range(i+1, n):
                    sum_xx = sum_xx + torch.norm(x[i]-x[j], p=1).to(device)
            sum_xx = sum_xx / (n * (n-1))

            sum_yy = torch.zeros(1).to(device)
            for i in range(y.shape[0]):
                for j in range(i+1, y.shape[0]):
                    sum_yy = sum_yy + torch.norm(y[i]-y[j], p=1).to(device)
            sum_yy = sum_yy / (y.shape[0] * (y.shape[0]-1))

            sum_xy = torch.zeros(1).to(device)
            for i in range(n):
                for j in range(y.shape[0]):
                    sum_xy = sum_xy + torch.norm(x[i]-y[j], p=1).to(device)
            sum_yy = sum_yy / (n * y.shape[0])
        else:
            qx = torch.sqrt(torch.clamp(x, eps, 1))
            qy = torch.sqrt(torch.clamp(y, eps, 1))
            xx = torch.matmul(qx, qx.t())
            yy = torch.matmul(qy, qy.t())
            xy = torch.matmul(qx, qy.t())

            def diffusion_kernel(a, tmpt):
                return torch.exp(-torch.acos(a).pow(2) / tmpt)

            off_diag = 1 - torch.eye(n).to(device)
            k_xx = diffusion_kernel(torch.clamp(xx, 0, 1-eps), t)
            k_yy = diffusion_kernel(torch.clamp(yy, 0, 1-eps), t)
            k_xy = diffusion_kernel(torch.clamp(xy, 0, 1-eps), t)
            sum_xx = (k_xx * off_diag).sum() / (n * (n-1))
            sum_yy = (k_yy * off_diag).sum() / (n * (n-1))
            sum_xy = 2 * k_xy.sum() / (n * n)
            # lamb = (4 * 3.1415926 * t) ** (-d/2)
        return (sum_xx + sum_yy - sum_xy)


    def contrastive_loss_cos(self, data, pos_data, neg_data, pos_num, neg_num, tau=1):
        cos = torch.nn.CosineSimilarity(dim=1)
        pos_data = torch.cat(pos_data, dim=0)
        neg_data = torch.cat(neg_data, dim=0)

        # normalize
        norm_value = torch.ones(data.shape[1]).to(self.device)
        # norm_value, _ = torch.max(torch.cat((data, pos_data, neg_data), dim=0), dim=0)
        # norm_value = torch.sum(torch.cat((data, pos_data, neg_data), dim=0), dim=0)
        data_ = data / norm_value.unsqueeze(0)
        pos_data_ = pos_data / norm_value.unsqueeze(0)
        neg_data_ = neg_data / norm_value.unsqueeze(0)

        data_for_pos_ = data_.repeat_interleave(pos_num, dim=0)
        data_for_neg_ = data_.repeat_interleave(neg_num, dim=0)

        pos_ = torch.exp(cos(data_for_pos_, pos_data_) / tau)
        neg_ = torch.exp(cos(data_for_neg_, neg_data_) / tau)
        pos_ = list(torch.split(pos_, list(pos_num), dim=0))
        neg_ = list(torch.split(neg_, list(neg_num), dim=0))

        loss = [(- torch.log(pos_i.mean() / (pos_i.mean() + neg_i.mean()))) for pos_i, neg_i in zip(pos_, neg_)]

        return sum(loss) / len(loss)

    def weighted_cos(self, A, B):
        """
        A: 64 * 1 * 9
        B: 64 * n * 9
        """
        w = F.softmax(self.celltype_weight, dim=0)
        dot_product = torch.sum(A * B * w, dim=2)
        norm_A = torch.norm(A, dim=2)
        norm_B = torch.norm(B, dim=2)
        weighted_cos = dot_product / (norm_A * norm_B + 1e-6)

        return weighted_cos

    def contrastive_loss_triplet(self, data, pos_data, neg_data, pos_num, neg_num, margin, device):
        start_time = time.time()
        pos_data_ = torch.split(pos_data, list(pos_num), dim=0)
        neg_data_ = torch.split(neg_data, list(neg_num), dim=0)

        pos_data__ = pad_sequence(pos_data_, batch_first=True)
        neg_data__ = pad_sequence(neg_data_, batch_first=True)

        data_for_pos_ = data.unsqueeze(1)
        data_for_neg_ = data.unsqueeze(1)
        pos_cos = self.weighted_cos(data_for_pos_, pos_data__)
        neg_cos = self.weighted_cos(data_for_neg_, neg_data__)

        loss = F.relu(- pos_cos.sum(dim=1) / pos_num + neg_cos.sum(dim=1) / neg_num + margin)

        end_time = time.time()

        return loss.mean()





import os 
import pandas as pd
import numpy as np
import rpy2.robjects as robjects
from rpy2.robjects import pandas2ri, conversion
import rpy2.robjects as ro


def run_gsea_with_processed_data(df_sorted, species='Homo sapiens', category='H', subcategory='NULL'):
    with conversion.localconverter(ro.default_converter + pandas2ri.converter):
        r_df = ro.conversion.py2rpy(df_sorted)
    robjects.globalenv['df_sorted'] = r_df
    if subcategory is None or subcategory == 'NULL':
        subcategory_code = "geneset <- msigdbr(species = species_name, category = category_name)"
    else:
        subcategory_code = f"geneset <- msigdbr(species = species_name, category = category_name, subcategory = '{subcategory}')"
    
    robjects.globalenv['species_name'] = species
    robjects.globalenv['category_name'] = category
    
    r_code = f"""
    library(clusterProfiler)
    library(msigdbr)
    
    gene_list <- df_sorted$logFC
    names(gene_list) <- df_sorted$SYMBOL
    
    {subcategory_code}
    TERM2GENE <- data.frame(term = geneset$gs_name, gene = geneset$gene_symbol)
    
    gsea_result <- GSEA(
      geneList = gene_list,
      TERM2GENE = TERM2GENE,
      pvalueCutoff = 0.05,
      pAdjustMethod = "BH",
      seed=0
    )
    
    results_df <- gsea_result@result
    """
    
    robjects.r(r_code)
    
    results_df = robjects.r['results_df']
    gsea_result = robjects.r['gsea_result']
    
    return pandas2ri.rpy2py(results_df), gsea_result


def create_gsea_plots_with_r(results_df, gsea_result, plot_dir):
    with conversion.localconverter(ro.default_converter + pandas2ri.converter):
        results_df = ro.conversion.py2rpy(results_df)
    robjects.globalenv['results_df'] = results_df
    robjects.globalenv['gsea_result'] = gsea_result
    robjects.globalenv['plot_dir'] = plot_dir
    os.makedirs(plot_dir, exist_ok=True)
    
    robjects.r("""
    library(ggplot2)
    library(enrichplot)
    library(clusterProfiler)
    """)
    
    robjects.r("""
    p1 <- gseaplot2(gsea_result, 3, color="blue", title="", base_size=13, ES_geom="line")
    """)
    
    robjects.r("""
    p <- ggplot(results_df, aes(x = NES, y = Description)) +
      geom_point(aes(size = setSize, color = p.adjust)) +
      scale_size_continuous(range = c(4,8)) +
      scale_color_gradient(low = "blue", high = "white") +
      labs(title = "Hallmark",
            x = "Normalized Enrichment Score (NES)",
            y = "",
            color = expression(~italic(P)~".adjust"),
            size = "Count") +
      theme_bw() + 
      theme(panel.grid = element_blank(), 
            axis.text.x = element_text(family='Arial', size = 16, color = "black"), 
            axis.text.y = element_text(family='Arial', size = 14, color = "black"), 
            axis.title = element_text(family='Arial', size = 14, color = "black")) +
      theme(plot.title = element_text(family='Arial', vjust = 0.1, hjust = 0.5, size=18, face='bold'),
            legend.title = element_text(family='Arial', size=16), 
            legend.text = element_text(size = rel(0.95)),
            legend.key.size = unit(20, "pt"))
    
    """)
    
    robjects.r("""
    ggsave(plot = p1, filename = file.path(plot_dir, "gsea_enrichment_pathway_plot.jpg"), width = 5, height = 5, dpi = 500)
    ggsave(plot = p, filename = file.path(plot_dir, "gsea_scatter_plot.jpg"), width = 10, height = 5, dpi = 500)
    """)


import os
import scanpy as sc
import pandas as pd
import numpy as np
import anndata as ad
import matplotlib.pyplot as plt
import rpy2.robjects as robjects
from rpy2.robjects import Formula
from rpy2.robjects.packages import importr
import math
from rpy2.robjects import pandas2ri, conversion
import rpy2.robjects as ro
mgcv = importr('mgcv')
stats = robjects.packages.importr('stats')


def bh_adjust(x, log=False):
    x = np.array(x)
    nai = np.where(~np.isnan(x))[0]
    ox = x.copy()
    x = x[nai]
    id = np.argsort(x)
    if log:
        q = x[id] + np.log(len(x) / np.arange(1, len(x) + 1))
    else:
        q = x[id] * len(x) / np.arange(1, len(x) + 1)
    a = np.minimum.accumulate(q[::-1])[::-1][np.argsort(id)]
    ox[nai] = a
    return ox


def find_overdispersed_genes(mat, gam_k=5, alpha=0.05, max_gene=1000):
    print('Gene filtering begin...')
    dfm = np.log(mat.mean(axis=0))
    dfv = np.log(mat.var(axis=0))
    df = pd.DataFrame({'m': dfm, 'v': dfv}, index=mat.columns)
    vi = np.isfinite(dfv)

    # Run R code in Python
    with conversion.localconverter(ro.default_converter + pandas2ri.converter):
        r_dataframe_vi = ro.conversion.py2rpy(df[vi])
    if len(vi) < gam_k * 1.5:
        gam_k = 1
    if gam_k < 2:
        m = stats.lm('v ~ m', data=r_dataframe_vi)
    else:
        fm = Formula('v ~ s(m, k = {})'.format(gam_k))
        m = mgcv.gam(fm, data=r_dataframe_vi)
    df['res'] = -np.inf
    df.loc[vi, 'res'] = stats.resid(m, type='response')

    df_res = pandas2ri.py2rpy(np.exp(df['res']))
    n_obs = mat.shape[0]

    lp = stats.pf(df_res, n_obs, n_obs, lower_tail=False, log_p=True)
    lpa = bh_adjust(lp, log=True)
    df['lp'] = lp
    df['lpa'] = lpa

    ods = df['lpa'] < np.log(alpha)
    print(f"{ods.sum()} overdispersed genes remaining.")

    if ods.sum() > max_gene:
        thres = sorted(df['lpa'].tolist())[max_gene]
        ods = df['lpa'] < thres
        print(f"select the top {ods.sum()}  most overdispersed genes by default.")
    df['ods'] = ods
    return df


def spatial_pp(df_expr, df_pos, barcodes, save_dir, min_counts=101, min_cells=None, max_gene_number=1000, filter_genes=True):
    """
    :param df_expr: (n_spots * n_genes), dataframe, with column names(gene names)
    :param df_pos: (n_spots * 2), dataframe
    :param barcodes: (n_spots,), numpy array
    :param save_dir: str
    """

    data_X = df_expr.to_numpy()
    data_obs = pd.DataFrame()
    data_obs['spot'] = barcodes
    data_obs['x'] = df_pos.iloc[:, 0]
    data_obs['y'] = df_pos.iloc[:, 1]
    data_var = pd.DataFrame(index=df_expr.columns)
    adata = ad.AnnData(data_X, obs=data_obs, var=data_var)

    print('Data preprocessing begin...')
    sc.pp.filter_cells(adata, min_counts=min_counts)
    print(f'Spots with no more than {min_counts - 1} gene counts are filtered out.')

    if filter_genes:
        sc.pp.filter_genes(adata, max_cells=len(adata.X) - 1)
        print('Genes present in all the spots are filtered out.')
        if min_cells == None:
            sc.pp.filter_genes(adata, min_cells=math.floor(0.05 * len(adata.X) + 1))
            print('Genes detected in fewer than 5% of spots are filtered out.')
        else:
            sc.pp.filter_genes(adata, min_cells=min_cells)
            print(f'Genes detected in fewer than {min_cells} spots are filtered out.')
        # find overdispersed genes
        mat = pd.DataFrame(adata.X, index=adata.obs['spot'], columns=adata.var.index)
        df = find_overdispersed_genes(mat, gam_k=5, alpha=0.05, max_gene=max_gene_number)
        data_index = np.nonzero((df['ods'] == 1).values)[0]
        data_select = adata.X[:, data_index]
        col_name = np.array(adata.var.index[df['ods']])
    else:
        data_select = adata.X
        col_name = adata.var.index.to_numpy()

    expr = pd.DataFrame(data_select, columns=col_name)
    id = pd.DataFrame(adata.obs['spot'].to_numpy(), columns=['spot_id'])
    df_pos_select = pd.DataFrame()
    df_pos_select['x'] = adata.obs['x']
    df_pos_select['y'] = adata.obs['y']
    df_pos_select.reset_index(drop=True, inplace=True)
    transcript_all = pd.concat([id, df_pos_select, expr], axis=1)

    # remove spots without gene expressions
    drop_index = []
    for i in range(len(transcript_all)):
        t = (transcript_all.iloc[i, 3:] > 0).sum()
        if t == 0:
            print([i, t])
            drop_index.append(i)
            
    os.makedirs(save_dir, exist_ok=True)
    transcript_all = transcript_all.drop(drop_index).reset_index(drop=True)
    transcript_all.to_csv(save_dir + '/spot_pos_and_expr.csv', index=False)

    print('\nData preprocessing end.')
    return transcript_all





import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="spatialsurf",
    version="1.15",
    author="Shuyu Liang",
    author_email="",
    description="A self-supervised deep learning method for reference-free deconvolution.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lllsssyyyy/SURF",
    packages=setuptools.find_packages(),
    install_requires=['pandas==1.5.3', 'numpy==1.23.0', 'scanpy==1.9.8', 'scipy==1.9.1',  'rpy2==3.6.1', 'tables', 'seaborn==0.13.2', 'matplotlib==3.7.1', 'pyarrow==14.0.1'],
    entry_points={
        'console_scripts': [
            'SURF=SURF:main'
        ],
    },
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
)